﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.1
@EndOfDesignText@
Sub Class_Globals
	Type person (idNumber As Long, initials As String, dob As String, gender As String, favColor As String)

	Private dataSource, alignments As List
	Private myColors, myColorsLong As B4XOrderedMap

	Dim peopleTable As B4XTable
	Private colorCol As B4XTableColumn
	Private Dialog As B4XDialog
	Private Base As B4XView
	Private InputNumberTemplate As B4XInputTemplate
	Private InputTextTemplate As B4XInputTemplate
	Private DateTemplate As B4XDateTemplate
	Private ColorTemplate As B4XColorTemplate
	Private ShortListTemplate As WiLListTemplate
	
	Private peopleSelections As B4XTableSelections
	Private BtnEditCell, btnUndo, btnUndoEdit As Button
	Private xBtnEditCell, xBtnUndo, xBtnUndoEdit As B4XView
	Private ExitNote As B4XView
	
	Private xui As XUI
	Private defaultTable As B4XTable
	
	Private mParent As B4XView
	#if B4A
		Private kb As IME			'need IME for soft keyboard in B4A
	#else if B4J
		Private fx As JFX			'need jFx for views in B4J
	#end if
	
	Private readOnly As B4XSet
	Private customEditType As Map
	Private LastEdit(3) As String
	Private LastDelete() As String
	Private LastDeletedRow As Long
	Private editMode As Boolean = False
	Private lineMode As Boolean = False
		
	Private buildingTable As Boolean = False
	Private genderList As List
End Sub

Public Sub Initialize (Parent As B4XView)
	Parent.LoadLayout("1")
	mParent = Parent
	Base = Parent

	readOnly.Initialize
	customEditType.Initialize
	dataSource.initialize
	alignments.Initialize

	prepareDialogs
	StartProgram
End Sub

Sub StartProgram
	DateTime.DateFormat = "yyyy-MM-dd"			'This is my preferred date format
	getColors									'A list of 144 "named" colors

	'dataSource is the source for the Table, it is generated here, but could be saved and retrieved from an external file
	'this means that the in-memory sql has a copy of all the data, but in a searchable, sortable, query-able form
	'this requires data modifications (cell edits and row deletions) to be done for both dataSource and the SQL set
	'inserting rows in the middle of the table requires reloading of the SQL (it happens very fast , even for 21500 rows on my tablet)
	dataSource.clear			
	CreatePeople(21500)

	'My preferences for putting as much specification in as little space as possible leads to ...
	'T=Text	D=Date I=Integer C=Color F=Float S=Spinner P=Password E=Email L=Link A=TextArea O=OptionList S=Signature
	'^=Sortable  ?= Searchable  !=Read Only <=LeftAlign >=RightAlign 
	
	peopleTable = makeTable ("ID~10~T^?!> | Initials~8~T<? | DOB~12~D^ | Age~5~I! | Gender~5~O^ | Favourite Color~20~C<^ | This Year's Birthday~40~T!<")
	Wait For (peopleTable.SetData(dataSource)) Complete

	'Prepare Selections object - a new Selections object should be created for each table - there is only one table here
	peopleSelections.Initialize(peopleTable)
	peopleSelections.Mode = peopleSelections.MODE_SINGLE_LINE_PERMANENT
	peopleSelections.SelectedTextColor = xui.Color_White
	peopleSelections.SelectionColor = xui.Color_ARGB(255, 0, 148, 255)
	peopleSelections.AutoRemoveInvisibleSelections = True

	Dim colorCol As B4XTableColumn = peopleTable.getColumn("Favourite Color")
	'Add image views to colorColumn, these are empty until the table is shown and are loaded on the fly to visible rows
	For i = 0 To colorCol.CellsLayouts.Size - 1
		Dim pnl As B4XView = colorCol.CellsLayouts.Get(i)
		Dim iv As ImageView
		iv.Initialize("")
		pnl.AddView(iv, .75*colorCol.Width, 5dip, 50dip, peopleTable.RowHeight - 10dip)
	Next
	
	'Record alignments - needs to be done after Table is ready - alignment is marked as tag on CellsLayout labels - forced on DataUpdate event
	For Each obj() As Object In alignments
		Dim aCol As B4XTableColumn = obj(0)
		Dim align As String = obj(1)
		For i = 1 To aCol.CellsLayouts.Size - 1
			Dim pnl As B4XView = aCol.CellsLayouts.Get(i)
			pnl.GetView(0).Tag = align
		Next
	Next

	'Switch side for searchfield for android (in this example, landscape mode is forced), only needed for B4A
	Dim pnlNav As B4XView = peopleTable.lblNumber.Parent
	#if B4A
		pnlNav.Left = 80%x
		pnlNav.Top = -5
		peopleTable.SearchField.mBase.left = 0
	#end if

	peopleTable.HighlightSearchResults = False
	upsizeSearchfield			'On my 10inch tablet the searchfiled font is too small
	
	'Add clickable panel for navigation page selection - a dialog is presented for a page choice - speeds up navigation
	Dim pageButton As Button
	pageButton.Initialize("pageButton")
	Dim pageButtonX As B4XView = pageButton
	pageButtonX.Color = xui.Color_Transparent
	pnlNav.AddView(pageButton, peopleTable.lblNumber.Left, peopleTable.lblNumber.Top, peopleTable.lblNumber.Width, peopleTable.lblNumber.Height)

	peopleTable.Refresh			'Render table with above changes

	'prepare controls shown below table
	xBtnUndo = btnUndo
	xBtnUndo.Top = mParent.Height - xBtnUndo.Height - 2dip
	xBtnUndo.BringToFront
	xBtnUndo.Enabled = False
	xBtnUndo.TextColor = xui.Color_ARGB(50, 0, 0, 0)

	xBtnUndoEdit = btnUndoEdit
	xBtnUndoEdit.Top = mParent.Height - xBtnUndoEdit.Height - 2dip
	xBtnUndoEdit.Enabled = False
	xBtnUndoEdit.TextColor = xui.Color_ARGB(50, 0, 0, 0)

	xBtnEditCell = BtnEditCell
	xBtnEditCell.Top = mParent.Height - xBtnEditCell.Height - 2dip
	xBtnEditCell.Enabled = True
	xBtnEditCell.TextColor = xui.Color_Black
	

	buildingTable = False			'The DataUpdate handler was disabled but can now start
	
	Dim tableSize(2) As Int = getTableSize(peopleTable)
	Log("Table size: rows = " & tableSize(0) & "  columns = " & tableSize(1))

	'show buttons after showing table
	xBtnEditCell.Visible = True
	xBtnUndo.Visible = True
	xBtnUndoEdit.Visible = True

	ExitNote.Top = mParent.Height - ExitNote.Height - 2dip
	ExitNote.BringToFront
	ExitNote.Visible = True
	peopleTable.mBase.Visible = True

	'move focus from buttons to table to be more neutral
	peopleTable.mBase.RequestFocus
End Sub

Sub makeTable(h As String) As B4XTable
	'multiple independent tables can be build and shown separately
	buildingTable = True		'disable DataUpdate handler
	
	#if B4A
		Dim numLines As Int = 14
	#else if B4J
		Dim numLines As Int = 16		'In my window on the desktop the table can have 2 more rows than on device
	#End If

	Dim TableBase As B4XView = xui.createPanel("tablebase")
	mParent.addView(TableBase, 0, 0, mParent.Width, mParent.Height)

	TableBase.LoadLayout("1")
	Dim table As B4XTable = defaultTable	'There is only one designer created defaultTable which is reloaded as for each new table
	table.mBase.Visible = False
	table.MaximumRowsPerPage = numLines
	table.BuildLayoutsCache(numLines)

	'Parse table specification	
	alignments.clear
	Dim header() As String = Regex.Split("\|", h)
	For Each s As String In header
		s = s.Trim
		Dim v() As String = Regex.Split("~", s)
		Dim colID As String = v(0)
		Dim aCol As B4XTableColumn
		Dim colType As Int = table.COLUMN_TYPE_TEXT
		aCol = table.AddColumn(colID, colType)
		Dim w As Float = v(1)
		aCol.Width = w * (table.mBase.width-.001) / 100
		
		'Parse header specs

		Dim specs As String = "T?"
		If v.Length>2 Then specs = v(2).toUpperCase
		Dim align As String = "CENTER"
		If specs.Contains("<") Then align = "LEFT"
		If specs.Contains(">") Then align = "RIGHT"

		If align <> "CENTER" Then alignments.add(Array(aCol, align))
		aCol.Searchable = False
		If specs.Contains("?") Then aCol.Searchable = True
		aCol.Sortable = False
		If specs.Contains("^") Then aCol.Sortable = True
		If specs.Contains("!") Then readOnly.Add(colID)
		If specs.Contains("T") Then
			customEditType.Put(colID, "AsText")
		Else if specs.Contains("D") Then
			customEditType.Put(colID, "AsDate")
		Else if specs.Contains("I") Then
			customEditType.Put(colID, "AsInteger")
		Else if specs.Contains("C") Then
			customEditType.Put(colID, "AsColor")
		Else if specs.Contains("F") Then
			customEditType.Put(colID, "AsFloat")
		Else if specs.Contains("S") Then
			customEditType.Put(colID, "AsSpinner")
		Else if specs.Contains("P") Then
			customEditType.Put(colID, "AsPassword")
		Else if specs.Contains("E") Then
			customEditType.Put(colID, "AsEmail")
		Else if specs.Contains("L") Then
			customEditType.Put(colID, "AsLink")
		Else if specs.Contains("A") Then
			customEditType.Put(colID, "AsLongText")
		Else if specs.Contains("O") Then
			customEditType.Put(colID, "AsList")
		Else if specs.Contains("S") Then
			customEditType.Put(colID, "Signature")
		End If
	Next
	Return table
End Sub

Sub defaultTable_DataUpdated
	If buildingTable Then Return
	Sleep(25)
	#if B4A
		If defaultTable.SearchField.Text = "" Then kb.HideKeyboard
	#End If
	For i = 0 To defaultTable.VisibleRowIds.Size - 1
		Dim RowId As Long = defaultTable.VisibleRowIds.Get(i)
		For j = 0 To defaultTable.Columns.size - 1
			Dim c As B4XTableColumn = defaultTable.Columns.Get(j)
			Dim pnl As B4XView = c.CellsLayouts.Get(i + 1)
			If RowId = 0 And i > 0 And c.Id = "Favourite Color" Then
				If pnl.NumberOfViews=2 And GetType(pnl.getView(1)).Contains("Image") Then
					Dim iv As B4XView = pnl.GetView(1)
					iv.SetBitmap(FontToBitmap("", False, 20, 0))
				End If
			End If
			If RowId > 0 Then
			Dim lbl As Label = pnl.GetView(0)
			Dim lt As String = lbl.Tag
				Dim xlbl As B4XView = lbl
				If lt = "LEFT" Then
					xlbl.SetTextAlignment("CENTER", "LEFT")
					lbl.left = 10dip				'Shifts label to simulate padding
				Else If lt = "RIGHT" Then
					xlbl.SetTextAlignment("CENTER", "RIGHT")
					lbl.left = -10dip				'Shifts label to simulate padding
				End If
				If c.Id = "Favourite Color" Then
					Dim iv As B4XView = pnl.GetView(1)
					iv.SetBitmap(FontToBitmap(Chr(0xF0C8), False, 20, convertCol(lbl.Text)))
					iv.BringToFront
				End If
			End If
		Next
	Next
	Dim pnl As B4XView = colorCol.CellsLayouts.Get(0)
	Dim iv As B4XView = pnl.GetView(1)
	iv.SetBitmap(FontToBitmap(Chr(0xF118), False, 20, xui.Color_Red))
End Sub

Sub getTableSize(table As B4XTable) As Int()
	Dim rs As ResultSet = table.sql1.ExecQuery("SELECT COUNT(*) FROM data")
	#if B4A
		rs.Position = 0
	#End If
	Dim w(2) As Int
	w(0) = rs.GetInt2(0)
	w(1) = table.Columns.Size
	Return w
End Sub

Sub hideTable(table As B4XTable)
	If table.IsInitialized Then
		ExitNote.Visible = False
		table.SearchField.Text = ""
		table.Clear
		Dim parent As B4XView = table.mBase.parent
		parent.RemoveAllViews
		parent.RemoveViewFromParent
		'hide table accessory buttons and panels with labels
		For Each bv As B4XView In mParent.GetAllViewsRecursive
			If GetType(bv).Contains("NonResize") Then
				bv.Visible = False
			Else If GetType(bv).Contains("Button") Then
				bv.Visible = False
			End If
		Next
		ExitApplication
	End If
End Sub

Sub defaultTable_CellClicked (columnId As String, rowId As Long)
	Dim tr As Map = defaultTable.GetRow(rowId)
	peopleSelections.CellClicked(columnId, rowId)		'Toggle selection - if on now off - if off now on
	If editMode = True And customEditType.ContainsKey(columnId) And readOnly.Contains(columnId) = False Then
		If peopleSelections.IsSelected = False Then peopleSelections.CellClicked(columnId, rowId)
		Dim cellValue As String = tr.Get(columnId)
		Dim ctype As String = customEditType.Get(columnId)
		LastEdit(0) = columnId
		LastEdit(1) = rowId
		LastEdit(2) = cellValue
		Dialog.Title = columnId
		Dialog.TitleBarHeight = 50dip
		Dialog.TitleBarColor = xui.Color_ARGB(255, 0, 148, 255)
		Select ctype
			Case "AsText"
				'Different responses may be needed for different columns, for example initials are AsText but need special treatment
				Select columnId
					Case "Initials": InputTextTemplate.lblTitle.Text = "periods/spaces are added automatically"
					Case Else: 'Dependent on application
				End Select
				InputTextTemplate.Text = cellValue
				Wait For (Dialog.ShowTemplate(InputTextTemplate, "Done", "", "CANCEL")) Complete (Result As Int)
				If Result = xui.DialogResponse_Positive Then
					Dim p As person = getPerson(rowId - 1)
					Select columnId
						Case "Initials": p.initials = formatInitials(InputTextTemplate.Text)
					End Select
					updateDataSource (p)
				Else
					BtnEditCell.Text = "Edit Cell"
					editMode = False
					peopleSelections.Clear
				End If
			Case "AsDate"
				If cellValue = "" Then cellValue = "2000-01-01"		'milestone date
				DateTemplate.Date = DateTime.DateParse(cellValue)
				Wait For (Dialog.ShowTemplate(DateTemplate, "", "", "CANCEL")) Complete (Result As Int)
				If Result = xui.DialogResponse_Positive Then
					Dim p As person = getPerson(rowId - 1)
					Select columnId
						Case "DOB":	p.dob = DateTime.Date(DateTemplate.Date)
					End Select
					updateDataSource (p)
				Else
					BtnEditCell.Text = "Edit Cell"
					editMode = False
					peopleSelections.Clear
				End If
			Case "AsColor"
				If cellValue = "" Then cellValue = "Olive Drab"
				ColorTemplate.SelectedHSVColor = ColorTemplate.ColorToHSV(convertCol(cellValue))
				Wait For (Dialog.ShowTemplate(ColorTemplate, "OK", "", "CANCEL")) Complete (Result As Int)
				If Result = xui.DialogResponse_Positive Then
					Dim p As person = getPerson(rowId - 1)
					Select columnId
						Case "Favourite Color": p.favColor = findClosestNamed(ColorTemplate.SelectedColor)
					End Select
					updateDataSource (p)
				Else
					BtnEditCell.Text = "Edit Cell"
					editMode = False
					peopleSelections.Clear
				End If
			Case "AsSpinner"
			Case "AsInteger"
			Case "AsFloat"
			Case "AsPassword"
			Case "AsEmail"
			Case "AsLink"
			Case "AsLongText"
			Case "AsList"
				ShortListTemplate.Initialize
				ShortListTemplate.Options = genderList
				Wait For (Dialog.ShowTemplate(ShortListTemplate, "", "", "Cancel")) Complete (Result As Int)
				If Result = xui.DialogResponse_Positive Then
					Dim p As person = getPerson(rowId - 1)
					Select columnId
						Case "Gender"
							Dim gend As String = ShortListTemplate.SelectedItem
							Select gend
								Case "Fluid": p.gender = "~"
								Case "Gender Queer": p.gender = "Q"
								Case "Gender Non-Binary": p.gender = "A"
								Case Else: p.gender = gend.charAt(0)
							End Select
					End Select
					updateDataSource (p)
				Else
					BtnEditCell.Text = "Edit Cell"
					editMode = False
					peopleSelections.Clear
				End If
			Case "AsSignature"
		End Select
	Else if lineMode = True And peopleSelections.IsSelected = False Then
		lineMode = False
		BtnEditCell.Text = "Edit Cell"
		editMode = False
		
		btnUndo.Text = "Undo last delete"
		btnUndo.Enabled = False
		xBtnUndo.TextColor = xui.Color_ARGB(50, 0, 0, 0)
		
		btnUndoEdit.Text = "Undo last edit"
		btnUndoEdit.Enabled = False
		xBtnUndoEdit.TextColor = xui.Color_ARGB(50, 0, 0, 0)
		peopleSelections.clear
	Else
		lineMode = True
		BtnEditCell.Text = "Delete Line"
		btnUndoEdit.Text = "Insert Line after"
		btnUndo.Text = "Insert Line before"
		btnUndoEdit.Enabled = True
		xBtnUndoEdit.TextColor = xui.Color_ARGB(200, 0, 0, 0)
		btnUndo.Enabled = True
		xBtnUndo.TextColor = xui.Color_ARGB(200, 0, 0, 0)
	End If
End Sub

Sub getPerson (where As Long) As person
	Dim theFacts() As String = dataSource.Get(where)
	Dim id As Long = theFacts(0).SubString(5)
	Return Createperson(id, theFacts(1), theFacts(2), theFacts(4), theFacts(5))
End Sub

Sub defaultTable_HeaderClicked (ColumnId As String)
	Sleep(25)
	If defaultTable.GetColumn(ColumnId).Sortable = False  Then
		For Each c As B4XTableColumn In defaultTable.Columns
			c.InternalSortMode = ""
		Next
		defaultTable.FirstRowIndex = 0
		defaultTable.Refresh
		Return
	End If
	defaultTable.lblSort.TextColor = xui.Color_Black
	upsizeSearchfield
End Sub

Sub upsizeSearchfield
	Dim searchParent As B4XView = defaultTable.SearchField.mBase.Parent
'	defaultTable.SearchField.mBase.Color = xui.Color_Red
	defaultTable.SearchField.TextField.Top = -15dip
	defaultTable.SearchField.TextField.Height = 85dip
'	searchParent.Color = xui.Color_Blue
	searchParent.Height = 100dip
		
	defaultTable.SearchField.SmallLabelTextSize = 16
	defaultTable.SearchField.LargeLabelTextSize = 24
	defaultTable.SearchField.TextField.TextSize = 24
			
	defaultTable.SearchField.TextField.TextColor = xui.Color_Black
	defaultTable.SearchField.HintColor = xui.Color_LightGray
	defaultTable.SearchField.lblClear.TextColor = xui.Color_Black
	defaultTable.SearchField.Update
End Sub

Sub pageButton_Click
	Dialog.Title = ""
	Dialog.TitleBarHeight = 50dip
	Dialog.TitleBarColor = xui.Color_ARGB(255, 0, 148, 255)
	InputNumberTemplate.Text = ""
	InputNumberTemplate.lblTitle.Text = "Page Number"
	Wait For (Dialog.ShowTemplate(InputNumberTemplate, "Done", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		If IsNumber(InputNumberTemplate.Text) Then
			Dim n As Int = InputNumberTemplate.Text
			defaultTable.CurrentPage = n
		End If
	End If
End Sub

Sub kb_HeightChanged (NewHeight As Int, OldHeight As Int)
	If defaultTable.IsInitialized Then
		defaultTable.mBase.Height = NewHeight - defaultTable.mBase.Top
		defaultTable.Refresh
	End If
End Sub

Public Sub Createperson (idNumber As Long, initials As String, dob As String, gender As String, favColor As String) As person
	Dim t1 As person
	t1.Initialize
	t1.idNumber = idNumber
	t1.initials = initials
	t1.dob = dob
	t1.gender = gender
	t1.favColor = favColor
	Return t1
End Sub

Sub exit_Click
	hideTable(peopleTable)
End Sub

Sub Application_Error (Error As Exception, StackTrace As String) As Boolean
	Return True
End Sub

Sub CreatePeople (n As Int)
	RndSeed(1357246813)
	Dim startDate As Long = DateTime.DateParse("1946-10-02")
	Dim endDate As Long = DateTime.DateParse("2002-01-01")
	Dim rangeDays As Int = (endDate-startDate)/DateTime.TicksPerDay
	For i = 1 To n
		Dim randomInitials As String = "ABCDEFGHIJKLMNOPQRSTUVWXYZ".CharAt(Rnd(0,26)) & ". " & "ABCDEFGHIJKLMNOPQRSTUVWXYZ".CharAt(Rnd(0,26)) & "."
		Dim randomDOB As Long = startDate + DateTime.TicksPerDay * Rnd(0, rangeDays)
		Dim randomGender As Int = Rnd(0,2)
		If Rnd(0, 1000) Mod 20 = 0 Then randomGender = 2 + Rnd(0,4)
		Dim randomFavCol As String = myColorsLong.Keys.Get(Rnd(0, myColorsLong.Size))
		Dim p As person = Createperson ((1000 + i), randomInitials, DateTime.Date(randomDOB), "MFABQ~".CharAt(randomGender), randomFavCol)
		dataSource.Add(Array As String("1023-" & p.idNumber, p.initials, p.dob, computeAge(p.dob), p.gender, p.favColor, computeBirthday(p.dob)))
	Next
End Sub

Sub BtnEditCell_Click
	If lineMode Then
		DeleteRow (defaultTable, peopleSelections.FirstSelectedRowId)
		lineMode = False
		editMode = False
		BtnEditCell.Text = "Edit Cell"
	Else
		peopleSelections.Clear
		If BtnEditCell.Text = "Edit Cell" Then
			BtnEditCell.Text = "Select Cell to Edit"
			editMode = True
		Else
			BtnEditCell.Text = "Edit Cell"
			editMode = False
		End If
	End If
End Sub

Sub btnUndo_Click
	If lineMode Then
		InsertRow (defaultTable, peopleSelections.FirstSelectedRowId, True)
	Else
		peopleSelections.Clear
		Dim saveCurrentPage As Long = defaultTable.CurrentPage
		dataSource.InsertAt(LastDeletedRow - 1, LastDelete)
		Wait For (defaultTable.SetData(dataSource)) Complete
		defaultTable.CurrentPage = saveCurrentPage
		xBtnUndo.Enabled = False
		xBtnUndo.TextColor = xui.Color_ARGB(50, 0, 0, 0)
	End If
End Sub

Sub btnUndoEdit_Click
	If lineMode Then
		InsertRow (defaultTable, peopleSelections.FirstSelectedRowId, False)
	Else
		Dim saveCurrentPage As Long = defaultTable.CurrentPage
		Dim p As person = getPerson(LastEdit(1)-1)
		'Only editable columns are considered here
		Select LastEdit(0)
			Case "Favourite Color": p.favColor = LastEdit(2)
			Case "DOB": p.dob = LastEdit(2)
			Case "Gender": p.gender = LastEdit(2)
			Case "Initials": p.initials = LastEdit(2)
		End Select
		updateDataSource(p)
		defaultTable.CurrentPage = saveCurrentPage
		xBtnUndoEdit.Enabled = False
		xBtnUndoEdit.TextColor = xui.Color_ARGB(50, 0, 0, 0)
	End If
End Sub

Sub DeleteRow (tbl As B4XTable, RowId As Long)
	Dim page As Int = tbl.CurrentPage
	LastDeletedRow = RowId
	LastDelete = dataSource.get(RowId - 1)
	dataSource.RemoveAt(RowId - 1)
	tbl.sql1.ExecNonQuery2("DELETE FROM data WHERE rowid = ?", Array (RowId))
	tbl.ClearDataView
	tbl.CurrentPage = page
	btnUndo.Text = "Undo last delete"
	btnUndo.Enabled = True
	xBtnUndo.TextColor = xui.Color_ARGB(200, 0, 0, 0)
	btnUndoEdit.Text = "Undo last edit"
	btnUndoEdit.Enabled = False
	xBtnUndoEdit.TextColor = xui.Color_ARGB(50, 0, 0, 0)
	peopleSelections.clear
End Sub

Sub InsertRow (tbl As B4XTable, RowId As Long, before As Boolean)
	If before Then RowId = RowId - 1
	Dim saveCurrentPage As Long = tbl.CurrentPage
	Dim p As person = Createperson ((1000 + dataSource.size + 1), "", "", "", "")
	dataSource.InsertAt(RowId, Array As String("1023-" & p.idNumber, p.initials, p.dob, "", p.gender, p.favColor, ""))
	peopleTable.SetData(dataSource)
	btnUndoEdit.Enabled = True
	xBtnUndoEdit.TextColor = xui.Color_ARGB(200, 0, 0, 0)
	BtnEditCell.Text = "Edit Cell"
	editMode = False
	peopleSelections.Clear
	tbl.CurrentPage = saveCurrentPage
	Sleep(100)
	
	peopleSelections.clear

	BtnEditCell.Text = "Select Cell to Edit"
	editMode = True
	lineMode = False
	btnUndo.Text = "Undo last delete"
	btnUndo.Enabled = False
	xBtnUndo.TextColor = xui.Color_ARGB(50, 0, 0, 0)
	
	btnUndoEdit.Text = "Undo last edit"
	btnUndoEdit.Enabled = False
	xBtnUndoEdit.TextColor = xui.Color_ARGB(50, 0, 0, 0)

End Sub

Private Sub MainForm_Resize (Width As Double, Height As Double)
End Sub

Sub updateTable(replaceRow() As String, where As Long)
	Dim args(replaceRow.Length+1) As String
	Dim sb As StringBuilder
	sb.Initialize
	sb.Append("UPDATE data SET ")
	For i = 0 To replaceRow.Length - 1
		Dim c As B4XTableColumn = defaultTable.VisibleColumns.Get(i)
		args(i) = replaceRow(i)
		If i = 0 Then
			sb.Append(c.SQLID).Append(" = ").Append("?")
		Else
			sb.Append(", ").Append(c.SQLID).Append(" = ").Append("?")
		End If
	Next
	args(replaceRow.Length) = where+1
	sb.Append(" WHERE rowid = ?")
	defaultTable.sql1.ExecNonQuery2(sb.toString, args)
	defaultTable.ClearDataView
End Sub

Sub addDerivedFields(p As person, where As Long)
	Dim zage As String = ""
	Dim zbd As String = ""
	If p.dob <> "" Then
		zage = computeAge(p.dob)
		zbd = computeBirthday(p.dob)
	End If
	Dim replaceRow() As String = Array As String("1023-" & p.idNumber, p.initials, p.dob, zage, p.gender, p.favColor, zbd)
	dataSource.set(where, replaceRow)
	updateTable(replaceRow, where)
End Sub

Sub updateDataSource(p As person)
	addDerivedFields(p, LastEdit(1) - 1)
	Dim moreToEdit() As String = dataSource.get(LastEdit(1)-1)
	Dim moreList As List: moreList.Initialize
	For i = 0 To moreToEdit.Length - 1
		If moreToEdit(i) = "" Then
			Dim tc As B4XTableColumn = peopleTable.Columns.Get(i)
			moreList.Add(tc.Id)
		End If
	Next
	If moreList.Size = 0 Then
		btnUndoEdit.Enabled = True
		xBtnUndoEdit.TextColor = xui.Color_ARGB(200, 0, 0, 0)
		BtnEditCell.Text = "Edit Cell"
		editMode = False
	End If
	peopleSelections.Clear
End Sub

Sub getColors
	myColors.Initialize
	myColorsLong.Initialize
	Dim col As List = File.ReadList(File.DirAssets, "colors.txt")
	For i = 0 To col.Size -1 Step 4
		Dim colNam As String = col.Get(i)
		Dim colVal As String = col.Get(i+2)
		colVal = colVal.Trim
		colVal = colVal.SubString2(1, colVal.Length-1)
		Dim w() As String = Regex.Split(",", colVal)
		Dim wn(3) As Int
		wn(0) = w(0)
		wn(1) = w(1)
		wn(2) = w(2)
		Dim v() As String = Regex.Split("\/", colNam)
		For Each s As String In v
			s = s.trim
			Dim g() As String = Regex.Split(" ", s)
			Dim sb As StringBuilder: sb.Initialize
			Dim sb2 As StringBuilder: sb2.Initialize
			For Each t As String In g
				t = t.trim
				sb.Append(t.ToUpperCase)
				sb2.Append(" " & t.SubString2(0,1).ToUpperCase & t.SubString(1))
			Next
			myColors.Put(sb.toString, wn)
			myColorsLong.Put((sb2.ToString).substring(1), wn)
		Next
	Next
End Sub

Sub convertCol(x As String) As Int
	x = x.toUpperCase
	x = x.Replace(" ", "")
	If x.IndexOf("RGB")=-1 Then
		Dim nam As String = x
		If myColors.ContainsKey(nam) Then
			Dim wn() As Int = myColors.Get(nam)
			Return xui.Color_RGB(wn(0), wn(1), wn(2))
		Else
			Return 0
		End If
	Else if x.IndexOf("ARGB")>-1 Then
		Dim colVal As String = x.trim.SubString(15)
		colVal = colVal.substring2(0,colVal.Length-1)
		Dim w() As String = Regex.Split(",", colVal)
		Dim wn(4) As Int
		wn(0) = w(0)
		wn(1) = w(1)
		wn(2) = w(2)
		wn(3) = w(3)
		Return xui.Color_ARGB(wn(0), wn(1), wn(2), wn(3))
	Else
		Dim colVal As String = x.trim.SubString(4)
		colVal = colVal.substring2(0,colVal.Length-1)
		Dim w() As String = Regex.Split(";", colVal)
		Dim wn(3) As Int
		wn(0) = w(0)
		wn(1) = w(1)
		wn(2) = w(2)
		Return xui.Color_RGB(wn(0), wn(1), wn(2))
	End If
End Sub

Sub findClosestNamed(clr As Int) As String
	Dim bc As ByteConverter
	Dim ix() As Byte = bc.IntsToBytes(Array As Int(clr))
	Dim ixint(3) As Int
	For j= 1 To 3
		ixint(j-1) = ix(j)
		If ix(j)<0 Then ixint(j-1) = 256 + ixint(j-1)
	Next
	Dim minDist As Int = 3*256
	Dim savekw As String = ""
	For Each kw As String In myColorsLong.Keys
		Dim wn() As Int = myColorsLong.Get(kw)
		Dim dist As Int = Abs(wn(0) - ixint(0)) + Abs(wn(1) - ixint(1)) + Abs(wn(2) - ixint(2))
		If dist<minDist Then
			minDist = dist
			savekw = kw
		End If
	Next
	Return savekw
End Sub

Public Sub FontToBitmap (text As String, IsMaterialIcons As Boolean, FontSize As Float, color As Int) As B4XBitmap
	Dim pnl As B4XView = xui.CreatePanel("")
	#if B4A
		Dim t As Typeface
		If IsMaterialIcons Then t = Typeface.MATERIALICONS Else t = Typeface.FONTAWESOME
	#else if B4J
		Dim t As Font
		If IsMaterialIcons Then t = fx.CreateMaterialIcons(20) Else t = fx.CreateFontAwesome(20)
	#end if
	pnl.SetLayoutAnimated(0, 0, 0, 32dip, 32dip)
	Dim fnt As B4XFont = xui.CreateFont(t, FontSize)

	Dim cvs1 As B4XCanvas
	cvs1.Initialize(pnl)
	Dim r As B4XRect = cvs1.MeasureText(text, fnt)
	Dim BaseLine As Int = cvs1.TargetRect.CenterY - r.Height / 2 - r.Top
	cvs1.DrawText(text, cvs1.TargetRect.CenterX, BaseLine, fnt, color, "CENTER")
	Dim b As B4XBitmap = cvs1.CreateBitmap
	cvs1.Release
	Return b
End Sub

Sub prepareDialogs
	#if B4A
		kb.Initialize("kb")
	#end if
	genderList = makeList(Array ("Male", "Female", "Gender Non-Binary", "Bi Gender", "Gender Queer", "Fluid"))
	Dim aLabel As Label
	aLabel.Initialize("exit")
	Dim ExitNote As B4XView = aLabel
	ExitNote.Font = xui.CreateMaterialIcons(35)
	ExitNote.Text = Chr(0xE879)
	ExitNote.Visible = False
	ExitNote.TextColor = xui.Color_Gray
	#if B4A
		ExitNote.TextSize = 25
		mParent.AddView (ExitNote, mParent.Width-150dip, 15dip, 50dip,30Dip)
	#else if B4J
		ExitNote.TextSize = 35
		mParent.AddView (ExitNote, mParent.Width-150dip, 0, 50dip,30Dip)
	#End If

	Dialog.Initialize (Base)
	Dialog.BorderWidth = 1
	Dialog.BorderCornersRadius = 5dip
	Dialog.BackgroundColor = xui.Color_White
	Dialog.Title = ""
	Dialog.TitleBarHeight = 40dip
	Dialog.ButtonsColor = xui.Color_LightGray
	Dialog.ButtonsTextColor = xui.Color_Black
	Dialog.BlurBackground = False
	Dialog.OverlayColor = xui.Color_ARGB(100, 255, 255, 255)

	DateTemplate.Initialize
	DateTemplate.MinYear = 1900
	DateTemplate.MaxYear = 3000
	DateTemplate.Date = DateTime.Now
	
	ColorTemplate.Initialize
	ShortListTemplate.Initialize

	InputNumberTemplate.Initialize
	InputNumberTemplate.lblTitle.TextColor = xui.Color_ARGB(255, 0, 148, 255)
	InputNumberTemplate.lblTitle.TextSize = 15
	InputNumberTemplate.TextField1.Height = 50dip
	InputNumberTemplate.TextField1.TextColor = xui.Color_Black
	InputNumberTemplate.TextField1.TextSize = 18
	InputNumberTemplate.ConfigureForNumbers(False, False)

	InputTextTemplate.Initialize
	InputTextTemplate.lblTitle.TextColor = xui.Color_ARGB(255, 0, 148, 255)
	InputTextTemplate.lblTitle.TextSize = 15
	InputTextTemplate.TextField1.Height = 50dip
	InputTextTemplate.TextField1.TextColor = xui.Color_Black
	InputTextTemplate.TextField1.TextSize = 18

	SetLightTheme
End Sub

Sub SetLightTheme
	Dialog.TitleBarColor = 0xFFFF7505
	Dialog.TitleBarHeight = 80dip
	Dim TextColor As Int = 0xFF5B5B5B
	Dialog.BackgroundColor = xui.Color_White
	Dialog.ButtonsColor = xui.Color_White
	Dialog.ButtonsTextColor = Dialog.TitleBarColor
	Dialog.BorderColor = xui.Color_Black
	
	DateTemplate.DaysInWeekColor = xui.Color_Black
	DateTemplate.SelectedColor = 0xFF39D7CE
	DateTemplate.HighlightedColor = 0xFF00CEFF
	DateTemplate.DaysInMonthColor = TextColor
	DateTemplate.lblMonth.TextColor = TextColor
	DateTemplate.lblYear.TextColor = TextColor
	DateTemplate.SelectedColor = 0xFFFFA761
	For Each x As B4XView In Array(DateTemplate.btnMonthLeft, DateTemplate.btnMonthRight, DateTemplate.btnYearLeft, DateTemplate.btnYearRight)
		x.Color = xui.Color_Transparent: x.TextColor = xui.Color_LightGray
	Next
End Sub

Sub computeBirthday(d As String) As String
	Dim currentYear As String = DateTime.getYear(DateTime.Now)
	Dim curBD As String = currentYear & d.SubString(4)
	Dim thisYearBD As Long
	'if born on feb 29 and current year is not a leap year then need to celebrate the birthday one day earlier!
	If curBD.EndsWith("-02-29") And DateTime.GetDayOfYear(DateTime.DateParse(currentYear & "-12-31")) = 365 Then
		thisYearBD = DateTime.DateParse(curBD.substring2(0,9) & "8")
	Else
		thisYearBD = DateTime.DateParse(curBD)
	End If

	Dim daysAgo As Int = DateUtils.PeriodBetweenInDays(thisYearBD, DateTime.Now).Days
	Dim bmonth As String = DateUtils.GetMonthName (thisYearBD)
	Dim weekday As String = DateUtils.GetDayOfWeekName (thisYearBD)
	Dim when As String = "Today!"
	If daysAgo > 0 Then when = daysAgo & " days ago"
	If daysAgo < 0 Then when = (-daysAgo) & " days from now"
	Return when & " - " & weekday & " " & bmonth & " " & DateTime.GetDayOfMonth(thisYearBD)
End Sub

Sub computeAge(d As String) As Int
	Dim currentYear As String = DateTime.getYear(DateTime.Now)
	Dim thisYearBD As Long = DateTime.DateParse(currentYear & d.SubString(4))
	Dim daysAgo As Int = DateUtils.PeriodBetweenInDays(thisYearBD, DateTime.Now).Days
	Dim age As Int = currentYear - DateTime.GetYear(DateTime.DateParse(d))
	If daysAgo < 0 Then age = age - 1
	Return age
End Sub

Sub makeList (items() As Object) As List
	Dim res As List: res.Initialize
	For Each obj As Object In items
		res.Add(obj)
	Next
	Return res
End Sub

Sub formatInitials(s As String) As String
	s = s.replace(".", "")
	s = s.replace(" ", "")
	Dim sb As StringBuilder: sb.Initialize
	For j = 0 To s.Length-1
		Dim t As String = s.CharAt(j)
		sb.Append(t.toUppercase & ". ")
	Next
	sb.remove(sb.Length-1, sb.Length)
	Return sb.toString
End Sub